@@ -70,6 +70,7 @@ class Options:
         # Parse out every option beginning with `python-gapic`
         opts: DefaultDict[str, List[str]] = defaultdict(list)
         for opt in opt_string.split(','):
+            opt = opt.strip()
             # Parse out the key and value.
             value = 'true'
             if '=' in opt:
@@ -92,11 +93,15 @@ class Options:
 
         # If templates are specified, one of the specified directories
         # may be our default; perform that replacement.
-        templates = opts.pop('templates', ['DEFAULT'])
-        while 'DEFAULT' in templates:
-            templates[templates.index('DEFAULT')] = os.path.realpath(
-                os.path.join(os.path.dirname(__file__), '..', 'templates'),
-            )
+        default_token = 'DEFAULT'
+        templates = opts.pop('templates', [default_token])
+        default_path = os.path.realpath(
+            os.path.join(os.path.dirname(__file__), '..', 'templates'),
+        )
+        templates = [
+            (default_path if path == default_token else path)
+            for path in templates
+        ]
 
         retry_cfg = None
         retry_paths = opts.pop('retry-config', None)
