@@ -19,19 +19,23 @@ through an :class:`~.API` object.
 
 import collections
 import dataclasses
+import os
 import sys
-from itertools import chain
-from typing import Callable, Dict, FrozenSet, Mapping, Sequence, Set, Tuple
+from typing import Callable, Container, Dict, FrozenSet, Mapping, Optional, Sequence, Set, Tuple
 
+from google.api_core import exceptions  # type: ignore
 from google.longrunning import operations_pb2  # type: ignore
 from google.protobuf import descriptor_pb2
 
+import grpc  # type: ignore
+
 from gapic.generator import options
 from gapic.schema import metadata
 from gapic.schema import wrappers
 from gapic.schema import naming as api_naming
 from gapic.utils import cached_property
 from gapic.utils import to_snake_case
+from gapic.utils import RESERVED_NAMES
 
 
 @dataclasses.dataclass(frozen=True)
@@ -51,10 +55,13 @@ class Proto:
         return getattr(self.file_pb2, name)
 
     @classmethod
-    def build(cls, file_descriptor: descriptor_pb2.FileDescriptorProto,
-              file_to_generate: bool, naming: api_naming.Naming,
-              prior_protos: Mapping[str, 'Proto'] = None,
-              ) -> 'Proto':
+    def build(
+        cls, file_descriptor: descriptor_pb2.FileDescriptorProto,
+        file_to_generate: bool, naming: api_naming.Naming,
+        opts: options.Options = options.Options(),
+        prior_protos: Mapping[str, 'Proto'] = None,
+        load_services: bool = True
+    ) -> 'Proto':
         """Build and return a Proto instance.
 
         Args:
@@ -66,12 +73,18 @@ class Proto:
                 with the API.
             prior_protos (~.Proto): Previous, already processed protos.
                 These are needed to look up messages in imported protos.
+            load_services (bool): Toggle whether the proto file should
+                load its services. Not doing so enables a two-pass fix for
+                LRO response and metadata types in certain situations.
         """
-        return _ProtoBuilder(file_descriptor,
-                             file_to_generate=file_to_generate,
-                             naming=naming,
-                             prior_protos=prior_protos or {},
-                             ).proto
+        return _ProtoBuilder(
+            file_descriptor,
+            file_to_generate=file_to_generate,
+            naming=naming,
+            opts=opts,
+            prior_protos=prior_protos or {},
+            load_services=load_services
+        ).proto
 
     @cached_property
     def enums(self) -> Mapping[str, wrappers.EnumType]:
@@ -84,10 +97,10 @@ class Proto:
     @cached_property
     def messages(self) -> Mapping[str, wrappers.MessageType]:
         """Return top-level messages on the proto."""
-        return collections.OrderedDict([
+        return collections.OrderedDict(
             (k, v) for k, v in self.all_messages.items()
             if not v.meta.address.parent
-        ])
+        )
 
     @property
     def module_name(self) -> str:
@@ -109,18 +122,21 @@ class Proto:
         # Add names of all enums, messages, and fields.
         answer: Set[str] = {e.name for e in self.all_enums.values()}
         for message in self.all_messages.values():
-            answer = answer.union({f.name for f in message.fields.values()})
+            answer.update(f.name for f in message.fields.values())
             answer.add(message.name)
 
         # Identify any import module names where the same module name is used
         # from distinct packages.
-        modules: Dict[str, Set[str]] = {}
-        for t in chain(*[m.field_types for m in self.all_messages.values()]):
-            modules.setdefault(t.ident.module, set())
-            modules[t.ident.module].add(t.ident.package)
-        for module_name, packages in modules.items():
-            if len(packages) > 1:
-                answer.add(module_name)
+        modules: Dict[str, Set[str]] = collections.defaultdict(set)
+        for m in self.all_messages.values():
+            for t in m.recursive_field_types:
+                modules[t.ident.module].add(t.ident.package)
+
+        answer.update(
+            module_name
+            for module_name, packages in modules.items()
+            if len(packages) > 1 or module_name in RESERVED_NAMES
+        )
 
         # Return the set of collision names.
         return frozenset(answer)
@@ -137,17 +153,18 @@ class Proto:
             for use in a ``from package import module`` type
             of statement.
         """
-        answer = set()
         self_reference = self.meta.address.python_import
-        for t in chain(*[m.field_types for m in self.messages.values()]):
-            # Add the appropriate Python import for the field.
+
+        answer = {
+            t.ident.python_import
+            for m in self.all_messages.values()
             # Sanity check: We do make sure that we are not trying to have
             # a module import itself.
-            if t.ident.python_import != self_reference:
-                answer.add(t.ident.python_import)
+            for t in m.field_types if t.ident.python_import != self_reference
+        }
 
         # Done; return the sorted sequence.
-        return tuple(sorted(list(answer)))
+        return tuple(sorted(answer))
 
     def disambiguate(self, string: str) -> str:
         """Return a disambiguated string for the context of this proto.
@@ -178,10 +195,13 @@ class API:
     subpackage_view: Tuple[str, ...] = dataclasses.field(default_factory=tuple)
 
     @classmethod
-    def build(cls,
-              file_descriptors: Sequence[descriptor_pb2.FileDescriptorProto],
-              package: str = '',
-              opts: options.Options = options.Options()) -> 'API':
+    def build(
+        cls,
+        file_descriptors: Sequence[descriptor_pb2.FileDescriptorProto],
+        package: str = '',
+        opts: options.Options = options.Options(),
+        prior_protos: Mapping[str, 'Proto'] = None,
+    ) -> 'API':
         """Build the internal API schema based on the request.
 
         Args:
@@ -193,6 +213,9 @@ class API:
                 Protos with packages outside this list are considered imports
                 rather than explicit targets.
             opts (~.options.Options): CLI options passed to the generator.
+            prior_protos (~.Proto): Previous, already processed protos.
+                These are needed to look up messages in imported protos.
+                Primarily used for testing.
         """
         # Save information about the overall naming for this API.
         naming = api_naming.Naming.build(*filter(
@@ -200,16 +223,58 @@ class API:
             file_descriptors,
         ), opts=opts)
 
+        def disambiguate_keyword_fname(
+                full_path: str,
+                visited_names: Container[str]) -> str:
+            path, fname = os.path.split(full_path)
+            name, ext = os.path.splitext(fname)
+            if name in RESERVED_NAMES or full_path in visited_names:
+                name += "_"
+                full_path = os.path.join(path, name + ext)
+                if full_path in visited_names:
+                    return disambiguate_keyword_fname(full_path, visited_names)
+
+            return full_path
+
         # Iterate over each FileDescriptorProto and fill out a Proto
         # object describing it, and save these to the instance.
-        protos: Dict[str, Proto] = {}
+        #
+        # The first pass gathers messages and enums but NOT services or methods.
+        # This is a workaround for a limitation in protobuf annotations for
+        # long running operations: the annotations are strings that reference
+        # message types but do not require a proto import.
+        # This hack attempts to address a common case where API authors,
+        # not wishing to generate an 'unused import' warning,
+        # don't import the proto file defining the real response or metadata
+        # type into the proto file that defines an LRO.
+        # We just load all the APIs types first and then
+        # load the services and methods with the full scope of types.
+        pre_protos: Dict[str, Proto] = dict(prior_protos or {})
         for fd in file_descriptors:
-            protos[fd.name] = _ProtoBuilder(
+            fd.name = disambiguate_keyword_fname(fd.name, pre_protos)
+            pre_protos[fd.name] = Proto.build(
                 file_descriptor=fd,
                 file_to_generate=fd.package.startswith(package),
                 naming=naming,
-                prior_protos=protos,
-            ).proto
+                opts=opts,
+                prior_protos=pre_protos,
+                # Ugly, ugly hack.
+                load_services=False,
+            )
+
+        # Second pass uses all the messages and enums defined in the entire API.
+        # This allows LRO returning methods to see all the types in the API,
+        # bypassing the above missing import problem.
+        protos: Dict[str, Proto] = {
+            name: Proto.build(
+                file_descriptor=proto.file_pb2,
+                file_to_generate=proto.file_to_generate,
+                naming=naming,
+                opts=opts,
+                prior_protos=pre_protos,
+            )
+            for name, proto in pre_protos.items()
+        }
 
         # Done; return the API.
         return cls(naming=naming, all_protos=protos)
@@ -274,6 +339,13 @@ class API:
                                                       )
         return answer
 
+    def requires_package(self, pkg: Tuple[str, ...]) -> bool:
+        return any(
+            message.ident.package == pkg
+            for proto in self.all_protos.values()
+            for message in proto.all_messages.values()
+        )
+
 
 class _ProtoBuilder:
     """A "builder class" for Proto objects.
@@ -291,16 +363,22 @@ class _ProtoBuilder:
     """
     EMPTY = descriptor_pb2.SourceCodeInfo.Location()
 
-    def __init__(self, file_descriptor: descriptor_pb2.FileDescriptorProto,
-                 file_to_generate: bool,
-                 naming: api_naming.Naming,
-                 prior_protos: Mapping[str, Proto] = None):
+    def __init__(
+        self,
+        file_descriptor: descriptor_pb2.FileDescriptorProto,
+        file_to_generate: bool,
+        naming: api_naming.Naming,
+        opts: options.Options = options.Options(),
+        prior_protos: Mapping[str, Proto] = None,
+        load_services: bool = True
+    ):
         self.proto_messages: Dict[str, wrappers.MessageType] = {}
         self.proto_enums: Dict[str, wrappers.EnumType] = {}
         self.proto_services: Dict[str, wrappers.Service] = {}
         self.file_descriptor = file_descriptor
         self.file_to_generate = file_to_generate
         self.prior_protos = prior_protos or {}
+        self.opts = opts
 
         # Iterate over the documentation and place it into a dictionary.
         #
@@ -346,19 +424,30 @@ class _ProtoBuilder:
         # In this situation, we would not have come across the message yet,
         # and the field would have its original textual reference to the
         # message (`type_name`) but not its resolved message wrapper.
-        for message in self.proto_messages.values():
-            for field in message.fields.values():
-                if field.type_name and not any((field.message, field.enum)):
-                    object.__setattr__(
-                        field, 'message',
-                        self.proto_messages[field.type_name.lstrip('.')],
-                    )
+        orphan_field_gen = (
+            (field.type_name.lstrip('.'), field)
+            for message in self.proto_messages.values()
+            for field in message.fields.values()
+            if field.type_name and not (field.message or field.enum)
+        )
+        for key, field in orphan_field_gen:
+            maybe_msg_type = self.proto_messages.get(key)
+            maybe_enum_type = self.proto_enums.get(key)
+            if maybe_msg_type:
+                object.__setattr__(field, 'message', maybe_msg_type)
+            elif maybe_enum_type:
+                object.__setattr__(field, 'enum', maybe_enum_type)
+            else:
+                raise TypeError(
+                    f"Unknown type referenced in "
+                    "{self.file_descriptor.name}: '{key}'"
+                )
 
         # Only generate the service if this is a target file to be generated.
         # This prevents us from generating common services (e.g. LRO) when
         # they are being used as an import just to get types declared in the
         # same files.
-        if file_to_generate:
+        if file_to_generate and load_services:
             self._load_children(file_descriptor.service, self._load_service,
                                 address=self.address, path=(6,))
         # TODO(lukesneeringer): oneofs are on path 7.
@@ -386,24 +475,24 @@ class _ProtoBuilder:
             return naive
 
         # Return a context-aware proto object.
-        # Note: The services bind to themselves, because services get their
-        # own output files.
-        return dataclasses.replace(naive,
-                                   all_enums=collections.OrderedDict([
-                                       (k, v.with_context(collisions=naive.names))
-                                       for k, v in naive.all_enums.items()
-                                   ]),
-                                   all_messages=collections.OrderedDict([
-                                       (k, v.with_context(collisions=naive.names))
-                                       for k, v in naive.all_messages.items()
-                                   ]),
-                                   services=collections.OrderedDict([
-                                       (k, v.with_context(collisions=v.names))
-                                       for k, v in naive.services.items()
-                                   ]),
-                                   meta=naive.meta.with_context(
-                                       collisions=naive.names),
-                                   )
+        return dataclasses.replace(
+            naive,
+            all_enums=collections.OrderedDict(
+                (k, v.with_context(collisions=naive.names))
+                for k, v in naive.all_enums.items()
+            ),
+            all_messages=collections.OrderedDict(
+                (k, v.with_context(collisions=naive.names))
+                for k, v in naive.all_messages.items()
+            ),
+            services=collections.OrderedDict(
+                # Note: services bind to themselves because services get their
+                # own output files.
+                (k, v.with_context(collisions=v.names))
+                for k, v in naive.services.items()
+            ),
+            meta=naive.meta.with_context(collisions=naive.names),
+        )
 
     @cached_property
     def api_enums(self) -> Mapping[str, wrappers.EnumType]:
@@ -492,17 +581,127 @@ class _ProtoBuilder:
         # Done; return the answer.
         return answer
 
+    def _get_retry_and_timeout(
+        self,
+        service_address: metadata.Address,
+        meth_pb: descriptor_pb2.MethodDescriptorProto
+    ) -> Tuple[Optional[wrappers.RetryInfo], Optional[float]]:
+        """Returns the retry and timeout configuration of a method if it exists.
+
+        Args:
+            service_address (~.metadata.Address): An address object for the
+                service, denoting the location of these methods.
+            meth_pb (~.descriptor_pb2.MethodDescriptorProto): A
+                protobuf method objects.
+
+        Returns:
+            Tuple[Optional[~.wrappers.RetryInfo], Optional[float]]: The retry
+                and timeout information for the method if it exists.
+        """
+
+        # If we got a gRPC service config, get the appropriate retry
+        # and timeout information from it.
+        retry = None
+        timeout = None
+
+        # This object should be a dictionary that conforms to the
+        # gRPC service config proto:
+        #   Repo: https://github.com/grpc/grpc-proto/
+        #   Filename: grpc/service_config/service_config.proto
+        #
+        # We only care about a small piece, so we are just leaving
+        # it as a dictionary and parsing accordingly.
+        if self.opts.retry:
+            # The gRPC service config uses a repeated `name` field
+            # with a particular format, which we match against.
+            # This defines the expected selector for *this* method.
+            selector = {
+                'service': '{package}.{service_name}'.format(
+                    package='.'.join(service_address.package),
+                    service_name=service_address.name,
+                ),
+                'method': meth_pb.name,
+            }
+
+            # Find the method config that applies to us, if any.
+            mc = next((c for c in self.opts.retry.get('methodConfig', [])
+                       if selector in c.get('name')), None)
+            if mc:
+                # Set the timeout according to this method config.
+                if mc.get('timeout'):
+                    timeout = self._to_float(mc['timeout'])
+
+                # Set the retry according to this method config.
+                if 'retryPolicy' in mc:
+                    r = mc['retryPolicy']
+                    retry = wrappers.RetryInfo(
+                        max_attempts=r.get('maxAttempts', 0),
+                        initial_backoff=self._to_float(
+                            r.get('initialBackoff', '0s'),
+                        ),
+                        max_backoff=self._to_float(r.get('maxBackoff', '0s')),
+                        backoff_multiplier=r.get('backoffMultiplier', 0.0),
+                        retryable_exceptions=frozenset(
+                            exceptions.exception_class_for_grpc_status(
+                                getattr(grpc.StatusCode, code),
+                            )
+                            for code in r.get('retryableStatusCodes', [])
+                        ),
+                    )
+
+        return retry, timeout
+
+    def _maybe_get_lro(
+        self,
+        service_address: metadata.Address,
+        meth_pb: descriptor_pb2.MethodDescriptorProto
+    ) -> Optional[wrappers.OperationInfo]:
+        """Determines whether a method is a Long Running Operation (aka LRO)
+               and, if it is, return an OperationInfo that includes the response
+               and metadata types.
+
+        Args:
+            service_address (~.metadata.Address): An address object for the
+                service, denoting the location of these methods.
+            meth_pb (~.descriptor_pb2.MethodDescriptorProto): A
+                protobuf method objects.
+
+        Returns:
+            Optional[~.wrappers.OperationInfo]: The info for the long-running
+                operation, if the passed method is an LRO.
+        """
+        lro = None
+
+        # If the output type is google.longrunning.Operation, we use
+        # a specialized object in its place.
+        if meth_pb.output_type.endswith('google.longrunning.Operation'):
+            op = meth_pb.options.Extensions[operations_pb2.operation_info]
+            if not op.response_type or not op.metadata_type:
+                raise TypeError(
+                    f'rpc {meth_pb.name} returns a google.longrunning.'
+                    'Operation, but is missing a response type or '
+                    'metadata type.',
+                )
+            response_key = service_address.resolve(op.response_type)
+            metadata_key = service_address.resolve(op.metadata_type)
+            lro = wrappers.OperationInfo(
+                response_type=self.api_messages[response_key],
+                metadata_type=self.api_messages[metadata_key],
+            )
+
+        return lro
+
     def _get_methods(self,
                      methods: Sequence[descriptor_pb2.MethodDescriptorProto],
-                     address: metadata.Address, path: Tuple[int, ...],
+                     service_address: metadata.Address, path: Tuple[int, ...],
                      ) -> Mapping[str, wrappers.Method]:
         """Return a dictionary of wrapped methods for the given service.
 
         Args:
             methods (Sequence[~.descriptor_pb2.MethodDescriptorProto]): A
                 sequence of protobuf method objects.
-            address (~.metadata.Address): An address object denoting the
-                location of these methods.
+            service_address (~.metadata.Address): An address object for the
+                service, denoting the location of these methods.
             path (Tuple[int]): The source location path thus far, as understood
                 by ``SourceCodeInfo.Location``.
 
@@ -512,38 +711,24 @@ class _ProtoBuilder:
         """
         # Iterate over the methods and collect them into a dictionary.
         answer: Dict[str, wrappers.Method] = collections.OrderedDict()
-        for meth_pb, i in zip(methods, range(0, sys.maxsize)):
-            lro = None
-
-            # If the output type is google.longrunning.Operation, we use
-            # a specialized object in its place.
-            if meth_pb.output_type.endswith('google.longrunning.Operation'):
-                op = meth_pb.options.Extensions[operations_pb2.operation_info]
-                if not op.response_type or not op.metadata_type:
-                    raise TypeError(
-                        f'rpc {meth_pb.name} returns a google.longrunning.'
-                        'Operation, but is missing a response type or '
-                        'metadata type.',
-                    )
-                lro = wrappers.OperationInfo(
-                    response_type=self.api_messages[address.resolve(
-                        op.response_type,
-                    )],
-                    metadata_type=self.api_messages[address.resolve(
-                        op.metadata_type,
-                    )],
-                )
+        for i, meth_pb in enumerate(methods):
+            retry, timeout = self._get_retry_and_timeout(
+                service_address,
+                meth_pb
+            )
 
             # Create the method wrapper object.
             answer[meth_pb.name] = wrappers.Method(
                 input=self.api_messages[meth_pb.input_type.lstrip('.')],
-                lro=lro,
+                lro=self._maybe_get_lro(service_address, meth_pb),
                 method_pb=meth_pb,
                 meta=metadata.Metadata(
-                    address=address.child(meth_pb.name, path + (i,)),
+                    address=service_address.child(meth_pb.name, path + (i,)),
                     documentation=self.docs.get(path + (i,), self.EMPTY),
                 ),
                 output=self.api_messages[meth_pb.output_type.lstrip('.')],
+                retry=retry,
+                timeout=timeout,
             )
 
         # Done; return the answer.
@@ -645,7 +830,7 @@ class _ProtoBuilder:
         # Put together a dictionary of the service's methods.
         methods = self._get_methods(
             service.method,
-            address=address,
+            service_address=address,
             path=path + (2,),
         )
 
@@ -659,3 +844,7 @@ class _ProtoBuilder:
             service_pb=service,
         )
         return self.proto_services[address.proto]
+
+    def _to_float(self, s: str) -> float:
+        """Convert a protobuf duration string (e.g. `"30s"`) to float."""
+        return int(s[:-1]) / 1e9 if s.endswith('n') else float(s[:-1])
