@@ -17,14 +17,24 @@ from unittest import mock
 
 import pytest
 
+from google.api import client_pb2
+from google.api_core import exceptions
 from google.longrunning import operations_pb2
 from google.protobuf import descriptor_pb2
 
+from gapic.generator import options
 from gapic.schema import api
 from gapic.schema import imp
 from gapic.schema import naming
 from gapic.schema import wrappers
 
+from test_utils.test_utils import (
+    make_field_pb2,
+    make_file_pb2,
+    make_message_pb2,
+    make_naming,
+)
+
 
 def test_api_build():
     # Put together a couple of minimal protos.
@@ -85,6 +95,9 @@ def test_api_build():
         imp.Import(package=('google', 'dep'), module='dep_pb2'),
     )
 
+    assert api_schema.requires_package(('google', 'example', 'v1'))
+    assert not api_schema.requires_package(('elgoog', 'example', 'v1'))
+
     # Establish that the subpackages work.
     assert 'common' in api_schema.subpackages
     sub = api_schema.subpackages['common']
@@ -137,6 +150,43 @@ def test_proto_names():
     assert proto.disambiguate('foo') == '_foo'
 
 
+def test_proto_keyword_fname():
+    # Protos with filenames that happen to be python keywords
+    # cannot be directly imported.
+    # Check that the file names are unspecialized when building the API object.
+    fd = (
+        make_file_pb2(
+            name='import.proto',
+            package='google.keywords.v1',
+            messages=(make_message_pb2(name='ImportRequest', fields=()),),
+        ),
+        make_file_pb2(
+            name='import_.proto',
+            package='google.keywords.v1',
+            messages=(make_message_pb2(name='ImportUnderRequest', fields=()),),
+        ),
+        make_file_pb2(
+            name='class_.proto',
+            package='google.keywords.v1',
+            messages=(make_message_pb2(name='ClassUnderRequest', fields=()),),
+        ),
+        make_file_pb2(
+            name='class.proto',
+            package='google.keywords.v1',
+            messages=(make_message_pb2(name='ClassRequest', fields=()),),
+        )
+    )
+
+    # We can't create new collisions, so check that renames cascade.
+    api_schema = api.API.build(fd, package='google.keywords.v1')
+    assert set(api_schema.protos.keys()) == {
+        'import_.proto',
+        'import__.proto',
+        'class_.proto',
+        'class__.proto',
+    }
+
+
 def test_proto_names_import_collision():
     # Put together a couple of minimal protos.
     fd = (
@@ -177,6 +227,110 @@ def test_proto_names_import_collision():
                            'other_message', 'primitive', 'spam'}
 
 
+def test_proto_names_import_collision_flattening():
+    lro_proto = api.Proto.build(make_file_pb2(
+        name='operations.proto', package='google.longrunning',
+        messages=(make_message_pb2(name='Operation'),),
+    ), file_to_generate=False, naming=make_naming())
+
+    fd = (
+        make_file_pb2(
+            name='mollusc.proto',
+            package='google.animalia.mollusca',
+            messages=(
+                make_message_pb2(name='Mollusc',),
+                make_message_pb2(name='MolluscResponse',),
+                make_message_pb2(name='MolluscMetadata',),
+            ),
+        ),
+        make_file_pb2(
+            name='squid.proto',
+            package='google.animalia.mollusca',
+            messages=(
+                make_message_pb2(
+                    name='IdentifySquidRequest',
+                    fields=(
+                        make_field_pb2(
+                            name='mollusc',
+                            number=1,
+                            type_name='.google.animalia.mollusca.Mollusc'
+                        ),
+                    ),
+                ),
+                make_message_pb2(
+                    name='IdentifySquidResponse',
+                    fields=(),
+                ),
+            ),
+            services=(
+                descriptor_pb2.ServiceDescriptorProto(
+                    name='SquidIdentificationService',
+                    method=(
+                        descriptor_pb2.MethodDescriptorProto(
+                            name='IdentifyMollusc',
+                            input_type='google.animalia.mollusca.IdentifySquidRequest',
+                            output_type='google.longrunning.Operation',
+                        ),
+                    ),
+                ),
+            ),
+        ),
+    )
+
+    method_options = fd[1].service[0].method[0].options
+    # Notice that a signature field collides with the name of an imported module
+    method_options.Extensions[client_pb2.method_signature].append('mollusc')
+    method_options.Extensions[operations_pb2.operation_info].MergeFrom(
+        operations_pb2.OperationInfo(
+            response_type='google.animalia.mollusca.MolluscResponse',
+            metadata_type='google.animalia.mollusca.MolluscMetadata',
+        )
+    )
+    api_schema = api.API.build(
+        fd,
+        package='google.animalia.mollusca',
+        prior_protos={
+            'google/longrunning/operations.proto': lro_proto,
+        }
+    )
+
+    actual_imports = {
+        ref_type.ident.python_import
+        for service in api_schema.services.values()
+        for method in service.methods.values()
+        for ref_type in method.ref_types
+    }
+
+    expected_imports = {
+        imp.Import(
+            package=('google', 'animalia', 'mollusca', 'types'),
+            module='mollusc',
+            alias='gam_mollusc',
+        ),
+        imp.Import(
+            package=('google', 'animalia', 'mollusca', 'types'),
+            module='squid',
+        ),
+        imp.Import(package=('google', 'api_core'), module='operation',),
+    }
+
+    assert expected_imports == actual_imports
+
+    method = (
+        api_schema
+        .services['google.animalia.mollusca.SquidIdentificationService']
+        .methods['IdentifyMollusc']
+    )
+
+    actual_response_import = method.lro.response_type.ident.python_import
+    expected_response_import = imp.Import(
+        package=('google', 'animalia', 'mollusca', 'types'),
+        module='mollusc',
+        alias='gam_mollusc',
+    )
+    assert actual_response_import == expected_response_import
+
+
 def test_proto_builder_constructor():
     sentinel_message = descriptor_pb2.DescriptorProto()
     sentinel_enum = descriptor_pb2.EnumDescriptorProto()
@@ -194,8 +348,8 @@ def test_proto_builder_constructor():
     # Test the load function.
     with mock.patch.object(api._ProtoBuilder, '_load_children') as lc:
         pb = api._ProtoBuilder(fdp,
-            file_to_generate=True,
-            naming=make_naming(),
+                               file_to_generate=True,
+                               naming=make_naming(),
                                )
 
         # There should be three total calls to load the different types
@@ -220,8 +374,8 @@ def test_proto_builder_constructor():
 
 def test_not_target_file():
     """Establish that services are not ignored for untargeted protos."""
-    message_pb = make_message_pb2(name='Foo',
-        fields=(make_field_pb2(name='bar', type=3, number=1),)
+    message_pb = make_message_pb2(
+        name='Foo', fields=(make_field_pb2(name='bar', type=3, number=1),)
     )
     service_pb = descriptor_pb2.ServiceDescriptorProto()
     fdp = make_file_pb2(messages=(message_pb,), services=(service_pb,))
@@ -237,8 +391,8 @@ def test_not_target_file():
 def test_messages():
     L = descriptor_pb2.SourceCodeInfo.Location
 
-    message_pb = make_message_pb2(name='Foo',
-        fields=(make_field_pb2(name='bar', type=3, number=1),)
+    message_pb = make_message_pb2(
+        name='Foo', fields=(make_field_pb2(name='bar', type=3, number=1),)
     )
     locations = (
         L(path=(4, 0), leading_comments='This is the Foo message.'),
@@ -269,7 +423,7 @@ def test_messages_reverse_declaration_order():
         make_message_pb2(name='Foo', fields=(
             make_field_pb2(name='bar', number=1,
                            type_name='.google.example.v3.Bar'),
-            ),
+        ),
         ),
         make_message_pb2(name='Bar'),
     )
@@ -294,7 +448,7 @@ def test_messages_recursive():
         make_message_pb2(name='Foo', fields=(
             make_field_pb2(name='foo', number=1,
                            type_name='.google.example.v3.Foo'),
-            ),
+        ),
         ),
     )
     fdp = make_file_pb2(
@@ -341,23 +495,192 @@ def test_messages_nested():
     assert bar not in proto.messages
 
 
+def test_out_of_order_enums():
+    # Enums can be referenced as field types before they
+    # are defined in the proto file.
+    # This happens when they're a nested type within a message.
+    messages = (
+        make_message_pb2(
+            name='Squid',
+            fields=(
+                make_field_pb2(
+                    name='base_color',
+                    type_name='google.mollusca.Chromatophore.Color',
+                    number=1,
+                ),
+            ),
+        ),
+        make_message_pb2(
+            name='Chromatophore',
+            enum_type=(
+                descriptor_pb2.EnumDescriptorProto(name='Color', value=()),
+            ),
+        )
+    )
+    fd = (
+        make_file_pb2(
+            name='squid.proto',
+            package='google.mollusca',
+            messages=messages,
+            services=(
+                descriptor_pb2.ServiceDescriptorProto(
+                    name='SquidService',
+                ),
+            ),
+        ),
+    )
+    api_schema = api.API.build(fd, package='google.mollusca')
+    field_type = (
+        api_schema
+        .messages['google.mollusca.Squid']
+        .fields['base_color']
+        .type
+    )
+    enum_type = api_schema.enums['google.mollusca.Chromatophore.Color']
+    assert field_type == enum_type
+
+
+def test_undefined_type():
+    fd = (
+        make_file_pb2(
+            name='mollusc.proto',
+            package='google.mollusca',
+            messages=(
+                make_message_pb2(
+                    name='Mollusc',
+                    fields=(
+                        make_field_pb2(
+                            name='class',
+                            type_name='google.mollusca.Class',
+                            number=1,
+                        ),
+                    )
+                ),
+            ),
+        ),
+    )
+    with pytest.raises(TypeError):
+        api.API.build(fd, package='google.mollusca')
+
+
+def test_python_modules_nested():
+    fd = (
+        make_file_pb2(
+            name='dep.proto',
+            package='google.dep',
+            messages=(make_message_pb2(name='ImportedMessage', fields=()),),
+        ),
+        make_file_pb2(
+            name='common.proto',
+            package='google.example.v1.common',
+            messages=(make_message_pb2(name='Bar'),),
+        ),
+        make_file_pb2(
+            name='foo.proto',
+            package='google.example.v1',
+            messages=(
+                make_message_pb2(
+                    name='GetFooRequest',
+                    fields=(
+                        make_field_pb2(name='primitive', number=2, type=1),
+                        make_field_pb2(
+                            name='foo',
+                            number=3,
+                            type=1,
+                            type_name='.google.example.v1.GetFooRequest.Foo',
+                        ),
+                    ),
+                    nested_type=(
+                        make_message_pb2(
+                            name='Foo',
+                            fields=(
+                                make_field_pb2(
+                                    name='imported_message',
+                                    number=1,
+                                    type_name='.google.dep.ImportedMessage'),
+                            ),
+                        ),
+                    ),
+                ),
+                make_message_pb2(
+                    name='GetFooResponse',
+                    fields=(
+                        make_field_pb2(
+                            name='foo',
+                            number=1,
+                            type_name='.google.example.v1.GetFooRequest.Foo',
+                        ),
+                    ),
+                ),
+            ),
+            services=(descriptor_pb2.ServiceDescriptorProto(
+                name='FooService',
+                method=(
+                    descriptor_pb2.MethodDescriptorProto(
+                        name='GetFoo',
+                        input_type='google.example.v1.GetFooRequest',
+                        output_type='google.example.v1.GetFooResponse',
+                    ),
+                ),
+            ),),
+        ),
+    )
+
+    api_schema = api.API.build(fd, package='google.example.v1')
+
+    assert api_schema.protos['foo.proto'].python_modules == (
+        imp.Import(package=('google', 'dep'), module='dep_pb2'),
+    )
+
+
 def test_services():
     L = descriptor_pb2.SourceCodeInfo.Location
 
+    # Make a silly helper method to not repeat some of the structure.
+    def _n(method_name: str):
+        return {
+            'service': 'google.example.v2.FooService',
+            'method': method_name,
+        }
+
+    # Set up retry information.
+    opts = options.Options(retry={'methodConfig': [
+        {'name': [_n('TimeoutableGetFoo')], 'timeout': '30s'},
+        {'name': [_n('RetryableGetFoo')], 'retryPolicy': {
+            'maxAttempts': 3,
+            'initialBackoff': '%dn' % 1e6,
+            'maxBackoff': '60s',
+            'backoffMultiplier': 1.5,
+            'retryableStatusCodes': ['UNAVAILABLE', 'ABORTED'],
+        }},
+    ]})
+
     # Set up messages for our RPC.
-    request_message_pb = make_message_pb2(name='GetFooRequest',
-        fields=(make_field_pb2(name='name', type=9, number=1),)
+    request_message_pb = make_message_pb2(
+        name='GetFooRequest', fields=(make_field_pb2(name='name', type=9, number=1),)
     )
     response_message_pb = make_message_pb2(name='GetFooResponse', fields=())
 
     # Set up the service with an RPC.
     service_pb = descriptor_pb2.ServiceDescriptorProto(
         name='FooService',
-        method=(descriptor_pb2.MethodDescriptorProto(
-            name='GetFoo',
-            input_type='google.example.v2.GetFooRequest',
-            output_type='google.example.v2.GetFooResponse',
-        ),),
+        method=(
+            descriptor_pb2.MethodDescriptorProto(
+                name='GetFoo',
+                input_type='google.example.v2.GetFooRequest',
+                output_type='google.example.v2.GetFooResponse',
+            ),
+            descriptor_pb2.MethodDescriptorProto(
+                name='TimeoutableGetFoo',
+                input_type='google.example.v2.GetFooRequest',
+                output_type='google.example.v2.GetFooResponse',
+            ),
+            descriptor_pb2.MethodDescriptorProto(
+                name='RetryableGetFoo',
+                input_type='google.example.v2.GetFooRequest',
+                output_type='google.example.v2.GetFooResponse',
+            ),
+        ),
     )
 
     # Fake-document our fake stuff.
@@ -370,6 +693,7 @@ def test_services():
 
     # Finally, set up the file that encompasses these.
     fdp = make_file_pb2(
+        name='test.proto',
         package='google.example.v2',
         messages=(request_message_pb, response_message_pb),
         services=(service_pb,),
@@ -377,14 +701,18 @@ def test_services():
     )
 
     # Make the proto object.
-    proto = api.Proto.build(fdp, file_to_generate=True, naming=make_naming())
+    proto = api.API.build(
+        [fdp],
+        'google.example.v2',
+        opts=opts,
+    ).protos['test.proto']
 
     # Establish that our data looks correct.
     assert len(proto.services) == 1
     assert len(proto.messages) == 2
     service = proto.services['google.example.v2.FooService']
     assert service.meta.doc == 'This is the FooService service.'
-    assert len(service.methods) == 1
+    assert len(service.methods) == 3
     method = service.methods['GetFoo']
     assert method.meta.doc == 'This is the GetFoo method.'
     assert isinstance(method.input, wrappers.MessageType)
@@ -393,6 +721,26 @@ def test_services():
     assert method.input.meta.doc == 'This is the GetFooRequest message.'
     assert method.output.name == 'GetFooResponse'
     assert method.output.meta.doc == 'This is the GetFooResponse message.'
+    assert not method.timeout
+    assert not method.retry
+
+    # Establish that the retry information on a timeout-able method also
+    # looks correct.
+    timeout_method = service.methods['TimeoutableGetFoo']
+    assert timeout_method.timeout == pytest.approx(30.0)
+    assert not timeout_method.retry
+
+    # Establish that the retry information on the retryable method also
+    # looks correct.
+    retry_method = service.methods['RetryableGetFoo']
+    assert retry_method.timeout is None
+    assert retry_method.retry.max_attempts == 3
+    assert retry_method.retry.initial_backoff == pytest.approx(0.001)
+    assert retry_method.retry.backoff_multiplier == pytest.approx(1.5)
+    assert retry_method.retry.max_backoff == pytest.approx(60.0)
+    assert retry_method.retry.retryable_exceptions == {
+        exceptions.ServiceUnavailable, exceptions.Aborted,
+    }
 
 
 def test_prior_protos():
@@ -537,6 +885,82 @@ def test_lro_missing_annotation():
         }, naming=make_naming())
 
 
+def test_cross_file_lro():
+    # Protobuf annotations for longrunning operations use strings to name types.
+    # As far as the protobuf compiler is concerned they don't reference the
+    # _types_ at all, so the corresponding proto file that owns the types
+    # does not need to be imported.
+    # This creates a potential issue when building rich structures around
+    # LRO returning methods. This test is intended to verify that the issue
+    # is handled correctly.
+
+    # Set up a prior proto that mimics google/protobuf/empty.proto
+    lro_proto = api.Proto.build(make_file_pb2(
+        name='operations.proto', package='google.longrunning',
+        messages=(make_message_pb2(name='Operation'),),
+    ), file_to_generate=False, naming=make_naming())
+
+    # Set up a method with LRO annotations.
+    method_pb2 = descriptor_pb2.MethodDescriptorProto(
+        name='AsyncDoThing',
+        input_type='google.example.v3.AsyncDoThingRequest',
+        output_type='google.longrunning.Operation',
+    )
+    method_pb2.options.Extensions[operations_pb2.operation_info].MergeFrom(
+        operations_pb2.OperationInfo(
+            response_type='google.example.v3.AsyncDoThingResponse',
+            metadata_type='google.example.v3.AsyncDoThingMetadata',
+        ),
+    )
+
+    # Set up the service with an RPC.
+    service_file = make_file_pb2(
+        name='service_file.proto',
+        package='google.example.v3',
+        messages=(
+            make_message_pb2(name='AsyncDoThingRequest', fields=()),
+        ),
+        services=(
+            descriptor_pb2.ServiceDescriptorProto(
+                name='LongRunningService',
+                method=(method_pb2,),
+            ),
+        )
+    )
+
+    # Set up the messages, including the annotated ones.
+    # This file is distinct and is not explicitly imported
+    # into the file that defines the service.
+    messages_file = make_file_pb2(
+        name='messages_file.proto',
+        package='google.example.v3',
+        messages=(
+            make_message_pb2(name='AsyncDoThingResponse', fields=()),
+            make_message_pb2(name='AsyncDoThingMetadata', fields=()),
+        ),
+    )
+
+    api_schema = api.API.build(
+        file_descriptors=(
+            service_file,
+            messages_file,
+        ),
+        package='google.example.v3',
+        prior_protos={'google/longrunning/operations.proto': lro_proto, },
+    )
+
+    method = (
+        api_schema.
+        all_protos['service_file.proto'].
+        services['google.example.v3.LongRunningService'].
+        methods['AsyncDoThing']
+    )
+
+    assert method.lro
+    assert method.lro.response_type.name == 'AsyncDoThingResponse'
+    assert method.lro.metadata_type.name == 'AsyncDoThingMetadata'
+
+
 def test_enums():
     L = descriptor_pb2.SourceCodeInfo.Location
     enum_pb = descriptor_pb2.EnumDescriptorProto(name='Silly', value=(
@@ -562,47 +986,3 @@ def test_enums():
     assert enum.values[1].meta.doc == 'This is the one value.'
     assert enum.values[2].name == 'THREE'
     assert enum.values[2].meta.doc == ''
-
-
-def make_file_pb2(name: str = 'my_proto.proto', package: str = 'example.v1', *,
-        messages: Sequence[descriptor_pb2.DescriptorProto] = (),
-        enums: Sequence[descriptor_pb2.EnumDescriptorProto] = (),
-        services: Sequence[descriptor_pb2.ServiceDescriptorProto] = (),
-        locations: Sequence[descriptor_pb2.SourceCodeInfo.Location] = (),
-                  ) -> descriptor_pb2.FileDescriptorProto:
-    return descriptor_pb2.FileDescriptorProto(
-        name=name,
-        package=package,
-        message_type=messages,
-        enum_type=enums,
-        service=services,
-        source_code_info=descriptor_pb2.SourceCodeInfo(location=locations),
-    )
-
-
-def make_message_pb2(
-        name: str,
-        fields: tuple = (),
-        **kwargs
-        ) -> descriptor_pb2.DescriptorProto:
-    return descriptor_pb2.DescriptorProto(name=name, field=fields, **kwargs)
-
-
-def make_field_pb2(name: str, number: int,
-        type: int = 11,  # 11 == message
-        type_name: str = None,
-                   ) -> descriptor_pb2.FieldDescriptorProto:
-    return descriptor_pb2.FieldDescriptorProto(
-        name=name,
-        number=number,
-        type=type,
-        type_name=type_name,
-    )
-
-
-def make_naming(**kwargs) -> naming.Naming:
-    kwargs.setdefault('name', 'Hatstand')
-    kwargs.setdefault('namespace', ('Google', 'Cloud'))
-    kwargs.setdefault('version', 'v1')
-    kwargs.setdefault('product_name', 'Hatstand')
-    return naming.Naming(**kwargs)
